@@ -1,4 +1,4 @@
-/* Copyright (c) 2000, 2010, Oracle and/or its affiliates. All rights reserved.
+/* Copyright (c) 2000, 2022, Oracle and/or its affiliates.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License, version 2.0,
@@ -37,7 +37,7 @@
     name		Name of file
 
   DESCRIPTION
-    The extension is defined as everything after the first extension character
+    The extension is defined as everything after the last extension character
     (normally '.') after the directory name.
 
   RETURN VALUES
@@ -51,7 +51,7 @@ char *fn_ext(const char *name)
   DBUG_ENTER("fn_ext");
   DBUG_PRINT("mfunkt",("name: '%s'",name));
 
-#if defined(FN_DEVCHAR) || defined(BASKSLASH_MBTAIL)
+#if defined(FN_DEVCHAR) || defined(_WIN32)
   {
     char buff[FN_REFLEN];
     size_t res_length;
@@ -61,6 +61,6 @@ char *fn_ext(const char *name)
   if (!(gpos= strrchr(name, FN_LIBCHAR)))
     gpos= name;
 #endif
-  pos=strchr(gpos,FN_EXTCHAR);
+  pos=strrchr(gpos,FN_EXTCHAR);
   DBUG_RETURN((char*) (pos ? pos : strend(gpos)));
 } /* fn_ext */
