@@ -679,7 +679,8 @@ bool check_grant_routine(THD *thd, ulong want_access,
                          TABLE_LIST *procs, bool is_proc, bool no_error);
 bool check_routine_level_acl(THD *thd, const char *db, const char *name,
                              bool is_proc);
-bool check_grant_db(THD *thd,const char *db);
+bool check_grant_db(THD *thd, const char *db,
+                    const bool check_table_grant = false);
 bool acl_check_proxy_grant_access(THD *thd, const char *host, const char *user,
                                   bool with_grant);
 void get_privilege_desc(char *to, uint max_length, ulong access);
